import logging
import signal
import resource
import asyncio
import uuid

from pyhap.accessory_driver import AccessoryDriver
from pyhap.loader import Loader
from serial import Serial
from serial.tools import list_ports
from RPi import GPIO
import requests
import json

from .config_files import HAP_PYTHON_CHARACTERISTICS_FILE, HAP_PYTHON_SERVICES_FILE, HAP_PYTHON_ACCESSORIES_FILE
from .hardware_interfaces.json import hardware_interface_decoder
from .hardware_interfaces.channels.json.channel_decoder import ChannelDecoder
from .globals import id_hardware_map, hardware_id_channels_map, id_channels_map, accessories, hub
import OpenHub.globals as glob

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

retry_strategy = Retry(
    total=10,
    backoff_factor=10,
    status_forcelist=[429, 500, 502, 503, 504],
    method_whitelist=["HEAD", "GET", "OPTIONS"]
)
adapter = HTTPAdapter(max_retries=retry_strategy)
http = requests.Session()
http.mount("https://", adapter)
http.mount("http://", adapter)

logging.basicConfig(level=logging.DEBUG, format="[%(module)s] %(message)s")

import os.path

file_path = '/home/openhubdaemon/openhub_serial.json'

logger = logging.getLogger(__name__)

if os.path.isfile(file_path):
    with open(file_path, 'r') as serial_file_json:
        serial_json = json.load(serial_file_json)
        hub_serial_no = str(serial_json['serial_no'])
        logger.info('SERIAL_NUMBER: ' + hub_serial_no)
else:
    with open(file_path, 'w') as serial_file_json:
        serial_json = {}
        hub_serial_no = str(uuid.uuid4())
        logger.info('SERIAL_NUMBER: ' + hub_serial_no)
        serial_json['serial_no'] = hub_serial_no
        json.dump(serial_json, serial_file_json)

def find_openhub_api_on_local_network():
    from .ip_resolver import map_network
    ips = map_network()
    logger.info('SERIAL_NUMBER: ' + hub_serial_no)

    for ip in ips:
        logger.debug('Trying ip: ' + str(ip))
        try:
            response = requests.get('http://'+str(ip)+':8000/openhubapi/about')
            if response.ok:
                api_ip = ip
                return api_ip
        except:
            continue

def check_if_api_has_hub_serial(openhub_api_ip):
    from .ip_resolver import get_ip
    ip = get_ip()
    response = requests.get('http://' + str(openhub_api_ip) + ':8000/hubs/'+str(hub_serial_no))
    if not response.ok:
        hub_dict = {}
        hub_dict['id'] = hub_serial_no
        hub_dict['ip'] = ip
        hub_dict['display_name'] = 'New ' + str(hub_serial_no)
        hub_dict['aid'] = 1
        response = requests.post('http://' + str(openhub_api_ip) + ':8000/hub/',json=hub_dict)

openhub_api_ip = None
while openhub_api_ip is None:
    try:
        openhub_api_ip = find_openhub_api_on_local_network()
    except:
        continue

check_if_api_has_hub_serial(openhub_api_ip)

def get_pico_ports():
    pts = list(list_ports.comports())
    COMs = []
    for pt in pts:
        if 'ACM' in pt[0]:
            serial_port = Serial(pt.device, 9600, timeout=1)
            COMs.append(serial_port)
    return COMs


def post_new_pico(pico_serial):
    pico_dict = {}
    pico_dict['id'] = pico_serial
    pico_dict['type'] = 'PiPico'
    pico_dict['hub'] = str(hub_serial_no)
    response = requests.post('http://' + str(openhub_api_ip) + ':8000/pipico/',json=pico_dict)


def setup_picos(COMS, hardwares, hardware_id_channels_map):
    for ser in COMS:
        command = "init"
        ser.write(command.encode('utf-8'))
        pico_response = ser.readline()
        pico_serial = pico_response[:-2].decode('utf8').replace("'", '"')
        print(pico_serial)

        if pico_serial not in hardwares.keys():
            post_new_pico(pico_serial)
            hardwares = load_hardware_config(hardwares)

        hardwares[pico_serial].set_serial_com(ser)

        channels_temp = []
        if pico_serial in hardware_id_channels_map.keys():
            for channel in hardware_id_channels_map[pico_serial]:
                response = requests.get('http://192.168.3.132:8000/channels/' + str(channel.serial_no) + '/io')
                data = response.json()
                pico_config_element = {}
                pico_config_element['serial_no'] = channel.serial_no
                for datum in data:
                    if 'label' in datum.keys() and datum['label'] is not None and 'pin' in datum.keys():
                        pico_config_element[datum['label']] = str(datum['pin'])
                    pico_config_element = {**pico_config_element, **datum}
                from OpenHub.hardware_interfaces.channels.pi_pico_analog import PiPicoAnalog
                from OpenHub.hardware_interfaces.channels.pi_pico_pump import PiPicoPump
                if channel.__class__.__name__ == PiPicoAnalog.__name__:
                    pico_config_element['type'] = 'sensor'
                elif channel.__class__.__name__ == PiPicoPump.__name__:
                    pico_config_element['type'] = 'pump'
                else:
                    pico_config_element['type'] = 'relay'
                pico_config_element['index'] = str(channel.channel_index)
                channels_temp.append(pico_config_element)

        pico_config = {'serial_no': pico_serial, 'no_channels': str(len(channels_temp)), 'channels': channels_temp}
        print((json.dumps(pico_config)))
        ser.write((json.dumps(pico_config) + '\n').encode())


def _gpio_setup(_cls, pin):
    if GPIO.mode() is None:
        GPIO.setmode(GPIO.BCM)
    GPIO.setup(pin, GPIO.OUT)


def load_hub_config(hub):
    from OpenHub.homekit_accessories.json import homekit_decoder

    response = http.get('http://192.168.3.132:8000/hubs/' + hub_serial_no)
    data = json.dumps(response.json())
    hub = json.loads(data, cls=homekit_decoder.HomekitDecoder)

    return hub


def load_hardware_config(hardware):
    response = http.get('http://192.168.3.132:8000/hub/' + hub_serial_no + '/hardwares')
    data = json.dumps(response.json())
    hardware_temp = json.loads(data, cls=hardware_interface_decoder.HardwareDecoder)
    for hard_t in hardware_temp:
        hardware['id'] = hard_t
    return hardware


def load_channels(channels, id_channels_map):
    for hard in id_hardware_map.values():
        response = requests.get('http://192.168.3.132:8000/hardwares/' + str(hard.serial_no) + '/channels')
        data = json.dumps(response.json())
        t = []
        for channel in response.json():
            t.append(json.loads(json.dumps(channel), cls=ChannelDecoder))
        channels[str(hard.serial_no)] = t
    for hard in id_hardware_map.values():
        for channel in channels[str(hard.serial_no)]:
            id_channels_map[channel.serial_no] = channel

    return channels, id_channels_map


def load_homekit_accessory_config(accessories):
    from OpenHub.homekit_accessories.json import homekit_decoder
    response = http.get('http://192.168.3.132:8000/hub/' + hub_serial_no + '/accessories')
    data = json.dumps(response.json())
    accessories_temp = json.loads(data, cls=homekit_decoder.HomekitDecoder)
    for accessory in accessories_temp:
        accessories[accessory.serial_no] = accessory
    return accessories


def setup_accessories(hub, accessories):
    for accessory in accessories.values():
        hub.add_accessory(accessory)
    glob.driver.add_accessory(accessory=hub)

    # driver.add_accessory(accessory)

#
# def load_calibration_config(data):
#     global calibration
#     calibration = json.loads(str(data), cls=raw_value_decoder.RawValueDecoder)

id_hardware_map = load_hardware_config(id_hardware_map)

hardware_id_channels_map, id_channels_map = load_channels(hardware_id_channels_map, id_channels_map)

COMS = get_pico_ports()
setup_picos(COMS, id_hardware_map, hardware_id_channels_map)


async def handle_client(reader, writer):
    request = None
    while request != 'quit':
        request = (await reader.readline())
        print(str(request))
        response = str(eval(request)) + '\n'
        print(str(response))
        writer.write(response.encode('utf8'))
        await writer.drain()
    writer.close()


async def run_server():
    server = await asyncio.start_server(handle_client, 'localhost', 8888)
    async with server:
        await server.serve_forever()


# asyncio.run(run_server())

#
# def get_bridge(driver):
#     bridge = GardenBridge.GardenBridge(driver)
#
#     return bridge
#
#
glob.loader = Loader(path_char=HAP_PYTHON_CHARACTERISTICS_FILE,
                     path_service=HAP_PYTHON_SERVICES_FILE)

glob.driver = AccessoryDriver(port=51826, persist_file=HAP_PYTHON_ACCESSORIES_FILE,
                              loader=glob.loader)

hub = load_hub_config(hub)

accessories = load_homekit_accessory_config(accessories)
setup_accessories(hub, accessories)

resource.setrlimit(resource.RLIMIT_NOFILE, (65536, 65536))

signal.signal(signal.SIGTERM, glob.driver.signal_handler)

glob.driver.start()



from hardware_interfaces.base_io.base_io_interface import BaseIOInterface

from adafruit_mcp3xxx.analog_in import AnalogIn
import logging
import uuid


class MCPAnalogIO(BaseIOInterface):
    logger = logging.getLogger(__name__)

    def __init__(self, mcp=None, channel_index=None, hardware_serial_no=None, serial_no=uuid.uuid4(),
                 *args, **kwargs):
        super(MCPAnalogIO, self).__init__(*args, **kwargs)
        self.analog_in = AnalogIn(mcp, channel_index)



import logging
import board
import busio

from hardware_interfaces.base_io.base_io_interface import BaseIOInterface


class SPI(BaseIOInterface):
    logger = logging.getLogger(__name__)

    def __init__(self, sck=board.SCK, miso=board.MISO, mosi=board.MOSI, *args, **kwargs):
        self.spi = busio.SPI(clock=sck, MISO=miso, MOSI=mosi)
        super(SPI, self).__init__(*args, **kwargs)



import digitalio
import logging
import board

from hardware_interfaces.base_io.base_io_interface import BaseIOInterface


class PiDigitalIO(BaseIOInterface):
    logger = logging.getLogger(__name__)

    def __init__(self, cs_pin=board.D5, *args, **kwargs):
        self.cs = digitalio.DigitalInOut(cs_pin)

        super(PiDigitalIO, self).__init__(*args, **kwargs)



import logging

from pyhap.const import CATEGORY_BRIDGE
from pyhap.accessory import Bridge
from pyhap.accessory import Accessory

from pyhap.const import STANDALONE_AID

from OpenHub.globals import driver

class Hub(Bridge):
    logger = logging.getLogger(__name__)
    category = CATEGORY_BRIDGE

    def __init__(self, serial_no=None, display_name=None, *args, **kwargs):
        self.serial_no = serial_no
        self.aid = STANDALONE_AID
        Bridge.__init__(self,driver=driver, display_name=display_name,
                         *args, **kwargs)
        self.add_garden_hub_service()

        # driver.add_accessory(self)


    #
    # def __init__(self, driver, **kwargs):
    #     super().__init__(driver, self.display_name)

    def add_info_service(self):
        """Helper method to add the required `AccessoryInformation` service.

        Called in `__init__` to be sure that it is the first service added.
        May be overridden.
        """
        serv_info = self.driver.loader.get_service("AccessoryInformation")
        serv_info.configure_char("Name", value=self.display_name)
        serv_info.configure_char("Manufacturer", value="BellyFrito")
        serv_info.configure_char("Model", value="GardenBridge")
        serv_info.configure_char("SerialNumber", value=self.serial_no)
        self.add_service(serv_info)

    def add_garden_hub_service(self):
        serv_info = self.driver.loader.get_service("OpenHub")
        serv_info.configure_char("IsConfigured", False)

        self.add_service(serv_info)

    @Accessory.run_at_interval(2)
    async def run(self):
        await super().run()



import asyncio

from OpenHub.homekit_accessories.homkit_sensor_interface import HomeKitSensorInterface
import logging
from OpenHub.globals import driver
from pyhap.const import CATEGORY_CAMERA
from pyhap.camera import Camera as PyHapCamera
from pyhap.camera import VIDEO_CODEC_PARAM_PROFILE_ID_TYPES, VIDEO_CODEC_PARAM_LEVEL_TYPES

logger = logging.getLogger(__name__)


class Camera(PyHapCamera):
    run_debug_message = "Camera State: "
    FFMPEG_CMD = (
        """raspivid -o - -t 0 -rot 180 -w {width} -h {height} -fps {fps} -b 2000000 | cvlc -vvv stream:///dev/stdin --sout '#rtp{{sdp=rtsp://{address}:{v_port}}}' :demux=h264""")

    category = CATEGORY_CAMERA

    def __init__(self, serial_no=None, display_name=None, *args, **kwargs):
        self.category = CATEGORY_CAMERA
        self.FFMPEG_CMD = (
            """raspivid -o - -t 0 -rot 180 -w {width} -h {height} -fps {fps} -b 2000000 | cvlc -vvv stream:///dev/stdin --sout '#rtp{{sdp=rtsp://{address}:{v_port}}}' :demux=h264""")

        options = {"video": {
            "codec": {
                "profiles": [
                    VIDEO_CODEC_PARAM_PROFILE_ID_TYPES["BASELINE"],
                ],
                "levels": [
                    VIDEO_CODEC_PARAM_LEVEL_TYPES['TYPE3_1'],
                ],
            },
            "resolutions": [
                [1280, 720, 30],
            ],
        }, "audio": {

        }, "address": "192.168.1.136", 'start_stream_cmd': self.FFMPEG_CMD, 'srtp': False}
        self.display_name = self.set_display_name(display_name)
        self.serial_no = serial_no

        super().__init__(options, driver=driver, display_name=self.display_name, *args, **kwargs)
        # super(HomeKitSensorInterface, self).__init__(serial_no=serial_no, display_name=display_name, *args,
        #                                              **kwargs)

    def set_display_name(self, display_name):
        if display_name is None:
            return "Camera"
        else:
            return display_name

    # def add_functional_service(self):
    #     return self.add_preload_service('C')
    #
    # def add_functional_service_characteristic(self):
    #     # return self.service.configure_char(
    #     #     'On', setter_callback=self.set_relay)
    #     pass
    #     # ### For client extensions ###

    # def get_snapshot(self, image_size):  # pylint: disable=unused-argument, no-self-use
    #     """Return a jpeg of a snapshot from the camera.
    #     Overwrite to implement getting snapshots from your camera.
    #     :param image_size: ``dict`` describing the requested image size. Contains the
    #         keys "image-width" and "image-height"
    #     """
    #     with open(os.path.join(RESOURCE_DIR, 'snapshot.jpg'), 'rb') as fp:
    #         return fp.read()



import logging
from abc import ABC
import json


class RawValueConverter(ABC):
    logger = logging.getLogger(__name__)

    calibration_directory = None
    homekit_accessory_serial_no = None

    def __init__(self):
        super().__init__()

    def add_accessory_serial_no(self, serial_no):
        if self.homekit_accessory_serial_no is None:
            self.homekit_accessory_serial_no = serial_no



import logging
from OpenHub.homekit_accessories.homkit_sensor_interface import HomeKitSensorInterface


class AirQuality(HomeKitSensorInterface):
    logger = logging.getLogger(__name__)
    run_debug_message = "Current Air Quality: "

    def __init__(self, serial_no=None, display_name=None, channel_interface_serial_no=None, *args, **kwargs):
        super().__init__(serial_no=serial_no, display_name=display_name,
                         channel_interface_serial_no=channel_interface_serial_no, *args, **kwargs)

    def set_display_name(self, display_name):
        if display_name is None:
            return "Air Quality"
        else:
            return display_name

    def add_functional_service(self):
        return self.add_preload_service('AirQualitySensor')

    def add_additional_services(self):
        self.pm25 = self.add_preload_service('PM2.5Density')
        self.pm100 = self.add_preload_service('PM10Density')

    def add_functional_service_characteristic(self):
        return self.service.get_characteristic('AirQuality')

    async def run(self):
        data = await self.channel.get_raw_data()
        self.logger.info(self.display_name + " Output: " + str(data))
        self.char.set_value(3)
        self.pm25.set_value(data)




def best_fit(X, Y):
    xbar = sum(X) / len(X)
    ybar = sum(Y) / len(Y)
    n = len(X)  # or len(Y)

    numer = sum([xi * yi for xi, yi in zip(X, Y)]) - n * xbar * ybar
    denum = sum([xi ** 2 for xi in X]) - n * xbar ** 2

    m = numer / denum
    b = ybar - m * xbar

    print('best fit line:\ny = {:.2f} + {:.2f}x'.format(b, m))

    return m, b



from abc import ABC, abstractmethod
import uuid
import logging


class BaseIOInterface(ABC):
    logger = logging.getLogger(__name__)

    def __init__(self, child_hardware_id=None, parent_hardware_id=None, *args, **kwargs):
        self.child_hardware_id = child_hardware_id
        self.parent_hardware_id = parent_hardware_id
